"""Client code"""
import asyncio
import logging
import sys
from datetime import datetime, timedelta
from contextlib import contextmanager
from aionursery import Nursery, MultiError

from . import (
    AnswerCodes,
    ArcamException,
    CommandCodes,
    CommandPacket,
    ConnectionFailed,
    NotConnectedException,
    ResponseException,
    ResponsePacket,
    _read_packet,
    _write_packet
)
from .utils import Throttle, async_retry

_LOGGER = logging.getLogger(__name__)
_REQUEST_TIMEOUT = timedelta(seconds=3)
_REQUEST_THROTTLE = 0.2

_HEARTBEAT_INTERVAL = timedelta(seconds=5)
_HEARTBEAT_TIMEOUT  = _HEARTBEAT_INTERVAL + _HEARTBEAT_INTERVAL

class Client:
    def __init__(self, host, port, loop=None) -> None:
        self._reader = None
        self._writer = None
        self._loop = loop if loop else asyncio.get_event_loop()
        self._task = None
        self._listen = set()
        self._host = host
        self._port = port
        self._throttle = Throttle(_REQUEST_THROTTLE)
        self._timestamp = datetime.now()

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def loop(self):
        return self._loop

    @contextmanager
    def listen(self, listener):
        self._listen.add(listener)
        yield self
        self._listen.remove(listener)

    async def _process_heartbeat(self, writer):
        while True:
            delay = self._timestamp + _HEARTBEAT_INTERVAL - datetime.now()
            if delay > timedelta():
                await asyncio.sleep(delay.total_seconds())
            else:
                _LOGGER.debug("Sending ping")
                await _write_packet(
                    writer,
                    CommandPacket(1, CommandCodes.POWER, bytes([0xF0]))
                )
                self._timestamp = datetime.now()

    async def _process_data(self, reader):
        try:
            while True:
                try:
                    packet = await asyncio.wait_for(
                        _read_packet(self._reader),
                        _HEARTBEAT_TIMEOUT.total_seconds()
                    )
                except asyncio.TimeoutError as exception:
                    _LOGGER.warning("Missed all pings")
                    raise ConnectionFailed() from exception

                if packet is None:
                    _LOGGER.info("Server disconnected")
                    return

                _LOGGER.debug("Packet received: %s", packet)
                for listener in self._listen:
                    listener(packet)
        finally:
            self._reader = None

    async def process(self):
        cancelled = set()
        async def cancelled_watcher():
            try:
                while True:
                    await asyncio.sleep(100)
            except asyncio.CancelledError:
                cancelled.add(True)

        async with Nursery() as nursery:
            nursery.start_soon(cancelled_watcher())
            nursery.start_soon(self._process_data(self._reader))
            nursery.start_soon(self._process_heartbeat(self._writer))

        if cancelled:
            raise asyncio.CancelledError

    @property
    def connected(self):
        return self._reader is not None and not self._reader.at_eof()

    @property
    def started(self):
        return self._writer is not None

    async def start(self):
        if self._writer:
            raise ArcamException("Already started")

        _LOGGER.debug("Connecting to %s:%d", self._host, self._port)
        try:
            self._reader, self._writer = await asyncio.open_connection(
                self._host, self._port, loop=self._loop)
        except ConnectionError as exception:
            raise ConnectionFailed() from exception
        except OSError as exception:
            raise ConnectionFailed() from exception
        _LOGGER.info("Connected to %s:%d", self._host, self._port)

    async def stop(self):
        if self._writer:
            try:
                _LOGGER.info("Disconnecting from %s:%d", self._host, self._port)
                self._writer.close()
                if sys.version_info >= (3, 7):
                    await self._writer.wait_closed()
            except (ConnectionError, OSError):
                pass
            finally:
                self._writer = None
                self._reader = None

    @async_retry(2, asyncio.TimeoutError)
    async def _request(self, request: CommandPacket):
        if not self._writer:
            raise NotConnectedException()
        writer = self._writer # keep copy around if stopped by another task
        future = asyncio.Future()

        def listen(response: ResponsePacket):
            if (response.zn == request.zn and
                    response.cc == request.cc):
                if not (future.cancelled() or future.done()):
                    future.set_result(response)

        await self._throttle.get()

        async def req():
            _LOGGER.debug("Requesting %s", request)
            with self.listen(listen):
                await _write_packet(writer, request)
                self._timestamp = datetime.now()
                return await future

        return await asyncio.wait_for(
            req(),
            _REQUEST_TIMEOUT.total_seconds())

    async def request(self, zn, cc, data):
        response = await self._request(CommandPacket(zn, cc, data))

        if response.ac == AnswerCodes.STATUS_UPDATE:
            return response.data

        raise ResponseException.from_response(response)

class ClientContext:
    def __init__(self, client: Client):
        self._client = client
        self._task = None

    async def __aenter__(self):
        await self._client.start()
        self._task = asyncio.ensure_future(
            self._client.process(),
            loop=self._client.loop
        )
        return self._client

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._task:
            self._task.cancel()
            try:
                await self._task
            except asyncio.CancelledError:
                pass
        await self._client.stop()



"""Zone state"""
import asyncio
import logging
from typing import List

from . import (
    AnswerCodes,
    CommandCodes,
    DecodeMode2CH,
    DecodeModeMCH,
    IncomingAudioConfig,
    IncomingAudioFormat,
    MenuCodes,
    NotConnectedException,
    RC5Codes,
    ResponseException,
    ResponsePacket,
    SourceCodes,
    SOURCECODE_TO_RC5CODE_ZONE1,
    SOURCECODE_TO_RC5CODE_ZONE2,
)
from .client import Client

_LOGGER = logging.getLogger(__name__)

class State():
    def __init__(self, client: Client, zn: int):
        self._zn = zn
        self._client = client
        self._state = dict()

    async def start(self):
        # pylint: disable=protected-access
        self._client._listen.add(self._listen)

    async def stop(self):
        # pylint: disable=protected-access
        self._client._listen.remove(self._listen)

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()

    def to_dict(self):
        return {
            'POWER': self.get_power(),
            'VOLUME': self.get_volume(),
            'SOURCE': self.get_source(),
            'MUTE': self.get_mute(),
            'MENU': self.get_menu(),
            'INCOMING_AUDIO_FORMAT': self.get_incoming_audio_format(),
            'DECODE_MODE_2CH': self.get_decode_mode_2ch(),
            'DECODE_MODE_MCH': self.get_decode_mode_mch(),
            'DAB_STATION': self.get_dab_station(),
            'DLS_PDT': self.get_dls_pdt(),
            'RDS_INFORMATION': self.get_rds_information(),
        }

    def __repr__(self):
        return "State ({})".format(self.to_dict())

    def _listen(self, packet: ResponsePacket):
        if packet.zn != self._zn:
            return

        if packet.ac == AnswerCodes.STATUS_UPDATE:
            self._state[packet.cc] = packet.data
        else:
            self._state[packet.cc] = None

    @property
    def zn(self):
        return self._zn

    @property
    def client(self):
        return self._client

    def get(self, cc):
        return self._state[cc]

    def get_incoming_audio_format(self):
        value = self._state.get(CommandCodes.INCOMING_AUDIO_FORMAT)
        if value is None:
            return None, None
        return (IncomingAudioFormat.from_int(value[0]),
                IncomingAudioConfig.from_int(value[1]))

    def get_decode_mode_2ch(self) -> DecodeMode2CH:
        value = self._state.get(CommandCodes.DECODE_MODE_STATUS_2CH)
        if value is None:
            return None
        return DecodeMode2CH.from_bytes(value)

    async def set_decode_mode_2ch(self, mode: DecodeMode2CH):
        if mode == DecodeMode2CH.STEREO:
            command = RC5Codes.STEREO
        elif mode == DecodeMode2CH.DOLBY_PLII_IIx_MOVIE:
            command = RC5Codes.DOLBY_PLII_IIx_MOVIE
        elif mode == DecodeMode2CH.DOLBY_PLII_IIx_MUSIC:
            command = RC5Codes.DOLBY_PLII_IIx_MUSIC
        elif mode == DecodeMode2CH.DOLBY_PLII_IIx_GAME:
            command = RC5Codes.DOLBY_PLII_IIx_GAME
        elif mode == DecodeMode2CH.DOLBY_PL:
            command = RC5Codes.DOLBY_PL
        elif mode == DecodeMode2CH.DTS_NEO_6_CINEMA:
            command = RC5Codes.DTS_NEO_6_CINEMA
        elif mode == DecodeMode2CH.DTS_NEO_6_MUSIC:
            command = RC5Codes.DTS_NEO_6_MUSIC
        elif mode == DecodeMode2CH.MCH_STEREO:
            command = RC5Codes.MCH_STEREO
        else:
            raise ValueError("Unkown mapping for mode {}".format(mode))

        await self._client.request(
            self._zn, CommandCodes.SIMULATE_RC5_IR_COMMAND, command.value)

    def get_decode_mode_mch(self) -> DecodeModeMCH:
        value = self._state.get(CommandCodes.DECODE_MODE_STATUS_MCH)
        if value is None:
            return None
        return DecodeModeMCH.from_bytes(value)

    async def set_decode_mode_mch(self, mode: DecodeModeMCH):
        if mode == DecodeModeMCH.STEREO_DOWNMIX:
            command = RC5Codes.STEREO
        elif mode == DecodeModeMCH.MULTI_CHANNEL:
            command = RC5Codes.MULTI_CHANNEL
        elif mode == DecodeModeMCH.DOLBY_D_EX_OR_DTS_ES:
            command = RC5Codes.DOLBY_D_EX
        elif mode == DecodeModeMCH.DOLBY_PLII_IIx_MOVIE:
            command = RC5Codes.DOLBY_PLII_IIx_MOVIE
        elif mode == DecodeModeMCH.DOLBY_PLII_IIx_MUSIC:
            command = RC5Codes.DOLBY_PLII_IIx_MUSIC
        else:
            raise ValueError("Unkown mapping for mode {}".format(mode))

        await self._client.request(
            self._zn, CommandCodes.SIMULATE_RC5_IR_COMMAND, command.value)

    def get_power(self):
        value = self._state.get(CommandCodes.POWER)
        if value is None:
            return None
        return int.from_bytes(value, 'big')

    async def set_power(self, power: bool) -> None:
        if power:
            if self._zn == 1:
                command = RC5Codes.POWER_ON
            else:
                command = RC5Codes.POWER_ON_ZONE2
        else:
            if self._zn == 1:
                command = RC5Codes.POWER_OFF
            else:
                command = RC5Codes.POWER_OFF_ZONE2

        await self._client.request(
            self._zn, CommandCodes.SIMULATE_RC5_IR_COMMAND, command.value)

    def get_menu(self) -> MenuCodes:
        value = self._state.get(CommandCodes.MENU)
        if value is None:
            return None
        return MenuCodes.from_bytes(value)

    def get_mute(self) -> bool:
        value = self._state.get(CommandCodes.MUTE)
        if value is None:
            return None
        return int.from_bytes(value, 'big') == 0

    async def set_mute(self, mute: bool) -> None:
        if mute:
            if self._zn == 1:
                command = RC5Codes.MUTE_ON
            else:
                command = RC5Codes.MUTE_ON_ZONE2
        else:
            if self._zn == 1:
                command = RC5Codes.MUTE_OFF
            else:
                command = RC5Codes.MUTE_OFF_ZONE2

        await self._client.request(
            self._zn, CommandCodes.SIMULATE_RC5_IR_COMMAND, command.value)

    def get_source(self) -> SourceCodes:
        value = self._state.get(CommandCodes.CURRENT_SOURCE)
        if value is None:
            return None
        return SourceCodes.from_int(
            int.from_bytes(value, 'big'))

    def get_source_list(self) -> List[SourceCodes]:
        if self._zn == 1:
            return list(SOURCECODE_TO_RC5CODE_ZONE1.keys())
        else:
            return list(SOURCECODE_TO_RC5CODE_ZONE2.keys())

    async def set_source(self, src: SourceCodes) -> None:

        if self._zn == 1:
            command = SOURCECODE_TO_RC5CODE_ZONE1[src]
        else:
            command = SOURCECODE_TO_RC5CODE_ZONE2[src]

        await self._client.request(
            self._zn, CommandCodes.SIMULATE_RC5_IR_COMMAND, command.value)

    def get_volume(self) -> int:
        value = self._state.get(CommandCodes.VOLUME)
        if value is None:
            return None
        return int.from_bytes(value, 'big')

    async def set_volume(self, volume: int) -> None:
        await self._client.request(
            self._zn, CommandCodes.VOLUME, bytes([volume]))

    async def inc_volume(self) -> None:
        if self._zn == 1:
            command = RC5Codes.INC_VOLUME
        else:
            command = RC5Codes.INC_VOLUME_ZONE2

        await self._client.request(
            self._zn, CommandCodes.SIMULATE_RC5_IR_COMMAND, command.value)

    async def dec_volume(self) -> None:
        if self._zn == 1:
            command = RC5Codes.DEC_VOLUME
        else:
            command = RC5Codes.DEC_VOLUME_ZONE2

        await self._client.request(
            self._zn, CommandCodes.SIMULATE_RC5_IR_COMMAND, command.value)

    def get_dab_station(self):
        value = self._state.get(CommandCodes.DAB_STATION)
        if value is None:
            return None
        return value.decode('utf8').rstrip()

    def get_dls_pdt(self):
        value = self._state.get(CommandCodes.DLS_PDT_INFO)
        if value is None:
            return None
        return value.decode('utf8').rstrip()

    def get_rds_information(self):
        value = self._state.get(CommandCodes.RDS_INFORMATION)
        if value is None:
            return None
        return value.decode('utf8').rstrip()

    async def update(self):
        async def _update(cc):
            try:
                data = await self._client.request(self._zn, cc, bytes([0xF0]))
                self._state[cc] = data
            except ResponseException as e:
                _LOGGER.debug("Response error skipping %s - %s", cc, e.ac)
                self._state[cc] = None
            except NotConnectedException as e:
                _LOGGER.debug("Not connected skipping %s", cc)
                self._state[cc] = None
            except asyncio.TimeoutError:
                _LOGGER.error("Timeout requesting %s", cc)

        if self._client.connected:
            await asyncio.wait([
                _update(CommandCodes.POWER),
                _update(CommandCodes.VOLUME),
                _update(CommandCodes.MUTE),
                _update(CommandCodes.CURRENT_SOURCE),
                _update(CommandCodes.MENU),
                _update(CommandCodes.DECODE_MODE_STATUS_2CH),
                _update(CommandCodes.DECODE_MODE_STATUS_MCH),
                _update(CommandCodes.INCOMING_AUDIO_FORMAT),
                _update(CommandCodes.DAB_STATION),
                _update(CommandCodes.DLS_PDT_INFO),
                _update(CommandCodes.RDS_INFORMATION),
            ])
        else:
            if self._state:
                self._state = dict()




import argparse
import asyncio
import logging
import sys

from . import CommandCodes, SourceCodes, IncomingAudioFormat, IncomingAudioConfig, DecodeMode2CH, DecodeModeMCH, RC5Codes, CommandNotRecognised, _LOGGER, ResponsePacket, AnswerCodes
from .client import Client, ClientContext
from .server import Server, ServerContext
from .state import State

# pylint: disable=invalid-name

def auto_int(x):
    return int(x, 0)

def auto_bytes(x):
    print(x)
    return bytes.decode(x)

parser = argparse.ArgumentParser(description='Communicate with arcam receivers.')
parser.add_argument('--verbose', action='store_true')

subparsers = parser.add_subparsers(dest="subcommand")

parser_state = subparsers.add_parser('state')
parser_state.add_argument('--host', required=True)
parser_state.add_argument('--port', default=50000)
parser_state.add_argument('--zone', default=1, type=int)
parser_state.add_argument('--volume', type=int)
parser_state.add_argument('--source', type=auto_int)
parser_state.add_argument('--monitor', action='store_true')

parser_client = subparsers.add_parser('client')
parser_client.add_argument('--host', required=True)
parser_client.add_argument('--port', default=50000)
parser_client.add_argument('--zone', default=1, type=int)
parser_client.add_argument('--command', type=auto_int)
parser_client.add_argument('--data', nargs='+', default=[0xF0], type=auto_int)

parser_server = subparsers.add_parser('server')
parser_server.add_argument('--host', default='localhost')
parser_server.add_argument('--port', default=50000)


async def run_client(args):
    client = Client(args.host, args.port)
    async with ClientContext(client):
        result = await client.request(args.zone, args.command, bytes(args.data))
        print(result)

async def run_state(args):
    client = Client(args.host, args.port)
    async with ClientContext(client):
        state = State(client, args.zone)

        if args.volume is not None:
            await state.set_volume(args.volume)

        if args.source is not None:
            await state.set_source(args.source)

        if args.monitor:
            async with state:
                prev = repr(state)
                await state.update()
                while client.connected:
                    curr = repr(state)
                    if prev != curr:
                        print(curr)
                        prev = curr
                    await asyncio.sleep(delay=1)
        else:
            await state.update()
            print(state)


async def run_server(args):
    class DummyServer(Server):

        def __init__(self, host, port):
            super().__init__(host, port)

            self._volume = bytes([10])
            self._source = bytes([SourceCodes.PVR])
            self._audio_format = bytes([IncomingAudioFormat.PCM, IncomingAudioConfig.STEREO_ONLY])
            self._decode_mode_2ch = bytes([DecodeMode2CH.DOLBY_PLII_IIx_MUSIC])
            self._decode_mode_mch = bytes([DecodeModeMCH.DOLBY_PLII_IIx_MUSIC])

            self.register_handler(0x01, CommandCodes.POWER, bytes([0xF0]), self.get_power)
            self.register_handler(0x01, CommandCodes.VOLUME, bytes([0xF0]), self.get_volume)
            self.register_handler(0x01, CommandCodes.VOLUME, None, self.set_volume)
            self.register_handler(0x01, CommandCodes.CURRENT_SOURCE, bytes([0xF0]), self.get_source)
            self.register_handler(0x01, CommandCodes.INCOMING_AUDIO_FORMAT, bytes([0xF0]), self.get_incoming_audio_format)
            self.register_handler(0x01, CommandCodes.DECODE_MODE_STATUS_2CH, bytes([0xF0]), self.get_decode_mode_2ch)
            self.register_handler(0x01, CommandCodes.DECODE_MODE_STATUS_MCH, bytes([0xF0]), self.get_decode_mode_mch)
            self.register_handler(0x01, CommandCodes.SIMULATE_RC5_IR_COMMAND, None, self.ir_command)

        def get_power(self, **kwargs):
            return bytes([1])

        def set_volume(self, data, **kwargs):
            self._volume = data
            return self._volume

        def get_volume(self, **kwargs):
            return self._volume

        def get_source(self, **kwargs):
            return self._source

        def set_source(self, data, **kwargs):
            self._source = data
            return self._source

        def ir_command(self, data, **kwargs):
            source_commands = {
                RC5Codes.SELECT_AUX.value: SourceCodes.AUX,
                RC5Codes.SELECT_AV.value: SourceCodes.AV,
                RC5Codes.SELECT_PVR.value: SourceCodes.PVR,
                RC5Codes.SELECT_DISPLAY.value: SourceCodes.DISPLAY,
            }
            source = source_commands.get(data)
            if source:
                self.set_source(bytes([source]))
                return [
                    ResponsePacket(
                        zn=0x01,
                        cc=CommandCodes.SIMULATE_RC5_IR_COMMAND,
                        ac=AnswerCodes.STATUS_UPDATE,
                        data=data
                    ),
                    ResponsePacket(
                        zn=0x01,
                        cc=CommandCodes.CURRENT_SOURCE,
                        ac=AnswerCodes.STATUS_UPDATE,
                        data=bytes([source])
                    )
                ]

            raise CommandNotRecognised()

        def get_decode_mode_2ch(self, **kwargs):
            return self._decode_mode_2ch

        def get_decode_mode_mch(self, **kwargs):
            return self._decode_mode_mch

        def get_incoming_audio_format(self, **kwargs):
            return self._audio_format

    server = DummyServer(args.host, args.port)
    async with ServerContext(server):
        while True:
            await asyncio.sleep(delay=1)

def main():

    args = parser.parse_args()

    if args.verbose:
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

        channel = logging.StreamHandler(sys.stdout)
        channel.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        channel.setFormatter(formatter)
        root.addHandler(channel)


    loop = asyncio.get_event_loop()

    if args.subcommand == 'client':
        loop.run_until_complete(run_client(args))
    elif args.subcommand == 'state':
        loop.run_until_complete(run_state(args))
    elif args.subcommand == 'server':
        loop.run_until_complete(run_server(args))



"""Fake server"""
import asyncio
import logging

from . import (
    AnswerCodes,
    CommandNotRecognised,
    ResponseException,
    ResponsePacket,
    _read_command_packet,
    _write_packet
)

_LOGGER = logging.getLogger(__name__)

class Server():
    def __init__(self, host: str, port: int) -> None:
        self._server = None  # type: Optional[asyncio.Server]
        self._host = host
        self._port = port
        self._handlers = dict()
        self._tasks = list()

    async def send_status(self, response):
        _LOGGER.debug("Client update %s -> %s", response)
        await _write_packet(response)

    async def process(self, reader, writer):
        _LOGGER.debug("Client connected")
        task = asyncio.Task.current_task()
        self._tasks.append(task)
        try:
            await self.process_runner(reader, writer)
        finally:
            _LOGGER.debug("Client disconnected")
            self._tasks.remove(task)

    async def process_runner(self, reader, writer):
        while True:
            request = await _read_command_packet(reader)
            if request is None:
                _LOGGER.debug("Client disconnected")
                return

            responses = await self.process_request(request)
            _LOGGER.debug("Client command %s -> %s", request, responses)
            for response in responses:
                await _write_packet(writer, response)

    async def process_request(self, request):
        handler = self._handlers.get((request.zn, request.cc, request.data))
        if handler is None:
            handler = self._handlers.get((request.zn, request.cc))

        try:
            if handler:
                data = handler(
                    zn=request.zn,
                    cc=request.cc,
                    data=request.data)

                if isinstance(data, bytes):
                    response = [
                        ResponsePacket(
                            request.zn,
                            request.cc,
                            AnswerCodes.STATUS_UPDATE,
                            data)
                    ]
                else:
                    response = data
            else:
                raise CommandNotRecognised()
        except ResponseException as e:
            response = [
                ResponsePacket(
                    request.zn,
                    request.cc,
                    e.ac,
                    e.data or bytes()
                )
            ]

        return response

    def register_handler(self, zn, cc, data, fun):
        if data:
            self._handlers[(zn, cc, data)] = fun
        else:
            self._handlers[(zn, cc)] = fun

    async def start(self):
        _LOGGER.debug("Starting server")
        self._server = await asyncio.start_server(
            self.process,
            self._host,
            self._port)
        return self

    async def stop(self):
        if self._server:
            _LOGGER.debug("Stopping server")
            self._server.close()
            await self._server.wait_closed()
            self._server = None
        
        _LOGGER.debug("Cancelling clients %s", self._tasks)
        for task in self._tasks:
            task.cancel()
        await asyncio.wait(self._tasks)


class ServerContext():
    def __init__(self, server: Server):
        self._server = server

    async def __aenter__(self):
        await self._server.start()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._server.stop()



import asyncio
import functools
import logging
from datetime import datetime, timedelta

_LOGGER = logging.getLogger(__name__)

def async_retry(attempts=2, allowed_exceptions=()):
    def decorator(f):
        @functools.wraps(f)
        async def wrapper(*args, **kwargs):
            attempt = attempts
            while True:
                attempt -= 1

                try:
                    return await f(*args, **kwargs)
                except allowed_exceptions:
                    if attempt == 0:
                        raise
                    _LOGGER.warning("Retrying: %s %s", f, args)

        return wrapper
    return decorator


class Throttle:
    def __init__(self, delay):
        self._timestamp = datetime.now()
        self._lock = asyncio.Lock()
        self._delay = timedelta(seconds=delay)

    async def get(self):
        async with self._lock:
            timestamp = datetime.now()
            delay = (self._timestamp - timestamp).total_seconds()
            if delay > 0:
                await asyncio.sleep(delay)
            self._timestamp = datetime.now() + self._delay



"""Arcam AV Control"""
import asyncio
import enum
import logging
import sys
from typing import Union

import attr

PROTOCOL_STR = b'\x21'
PROTOCOL_ETR = b'\x0D'
PROTOCOL_EOF = b''

_LOGGER = logging.getLogger(__name__)
_WRITE_TIMEOUT = 3
_READ_TIMEOUT = 3

class ArcamException(Exception):
    pass

class ConnectionFailed(ArcamException):
    pass

class NotConnectedException(ArcamException):
    pass

class ResponseException(ArcamException):
    def __init__(self, ac=None, zn=None, cc=None, data=None):
        self.ac = ac
        self.zn = zn
        self.cc = cc
        self.data = data
        super().__init__("'ac':{}, 'zn':{}, 'cc':{}, 'data':{}".format(
            ac, zn, cc, data
        ))

    @staticmethod
    def from_response(response: 'ResponsePacket'):
        kwargs = {
            'zn': response.zn,
            'cc': response.cc,
            'data': response.data
        }
        if response.ac == AnswerCodes.ZONE_INVALID:
            return InvalidZoneException(**kwargs)
        elif response.ac == AnswerCodes.COMMAND_NOT_RECOGNISED:
            return CommandNotRecognised(**kwargs)
        elif response.ac == AnswerCodes.PARAMETER_NOT_RECOGNISED:
            return ParameterNotRecognised(**kwargs)
        elif response.ac == AnswerCodes.COMMAND_INVALID_AT_THIS_TIME:
            return CommandInvalidAtThisTime(**kwargs)
        elif response.ac == AnswerCodes.INVALID_DATA_LENGTH:
            return InvalidDataLength(**kwargs)
        else:
            return ResponseException(ac=response.ac, **kwargs)

class InvalidZoneException(ResponseException):
    def __init__(self, zn=None, cc=None, data=None):
        super().__init__(ac=AnswerCodes.ZONE_INVALID,
                         zn=zn, cc=cc, data=data)

class CommandNotRecognised(ResponseException):
    def __init__(self, zn=None, cc=None, data=None):
        super().__init__(ac=AnswerCodes.COMMAND_NOT_RECOGNISED,
                         zn=zn, cc=cc, data=data)

class ParameterNotRecognised(ResponseException):
    def __init__(self, zn=None, cc=None, data=None):
        super().__init__(ac=AnswerCodes.PARAMETER_NOT_RECOGNISED,
                         zn=zn, cc=cc, data=data)

class CommandInvalidAtThisTime(ResponseException):
    def __init__(self, zn=None, cc=None, data=None):
        super().__init__(ac=AnswerCodes.COMMAND_INVALID_AT_THIS_TIME,
                         zn=zn, cc=cc, data=data)

class InvalidDataLength(ResponseException):
    def __init__(self, zn=None, cc=None, data=None):
        super().__init__(ac=AnswerCodes.INVALID_DATA_LENGTH,
                         zn=zn, cc=cc, data=data)

class InvalidPacket(ArcamException):
    pass

class AnswerCodes(enum.IntEnum):
    STATUS_UPDATE = 0x00
    ZONE_INVALID = 0x82
    COMMAND_NOT_RECOGNISED = 0x83
    PARAMETER_NOT_RECOGNISED = 0x84
    COMMAND_INVALID_AT_THIS_TIME = 0x85
    INVALID_DATA_LENGTH = 0x86

    @staticmethod
    def from_int(value: int):
        try:
            return AnswerCodes(value)
        except ValueError:
            return value


class CommandCodes(enum.IntEnum):
    # System Commands
    POWER = 0x00
    DISPLAY_BRIGHTNESS = 0x01
    HEADPHONES = 0x02
    FMGENRE = 0x03
    SOFTWARE_VERSION = 0x04
    RESTORE_FACTORY_DEFAULT = 0x05
    SAVE_RESTORE_COPY_OF_SETTINGS = 0x06
    SIMULATE_RC5_IR_COMMAND = 0x08
    DISPLAY_INFORMATION_TYPE = 0x09
    CURRENT_SOURCE = 0x1D  # Request
    HEADPHONES_OVERRIDE = 0x1F


    # Input Commands
    VIDEO_SELECTION = 0x0A
    SELECT_ANALOG_DIGITAL = 0x0B
    VIDEO_INPUT_TYPE = 0x0C


    # Output Commands
    VOLUME = 0x0D  # Set/Request
    MUTE = 0x0E  # Request
    DIRECT_MODE_STATUS = 0x0F  # Request
    DECODE_MODE_STATUS_2CH = 0x10  # Request
    DECODE_MODE_STATUS_MCH = 0x11  # Request
    RDS_INFORMATION = 0x12  # Request
    VIDEO_OUTPUT_RESOLUTION = 0x13 # Set/Request


    # Menu Command
    MENU = 0x14  # Request
    TUNER_PRESET = 0x15  # Set/Request
    TUNE = 0x16  # Set/Request
    DAB_STATION = 0x18  # Set/Request
    DAB_PROGRAM_TYPE_CATEGORY = 0x19  # Set/Request
    DLS_PDT_INFO = 0x1A  # Request
    PRESET_DETAIL = 0x1B # Request


    # Network Command


    # Setup
    TREBLE_EQUALIZATION = 0x35
    BASS_EQUALIZATION = 0x36
    ROOM_EQUALIZATION = 0x37
    DOLBY_VOLUME = 0x38
    DOLBY_LEVELER = 0x39
    DOLBY_VOLUME_CALIBRATION_OFFSET = 0x3A
    BALANCE = 0x3B

    DOLBY_PLII_X_MUSIC_DIMENSION = 0x3C
    DOLBY_PLII_X_MUSIC_CENTRE_WIDTH = 0x3D
    DOLBY_PLII_X_MUSIC_PANORAMA = 0x3E
    SUBWOOFER_TRIM = 0x3F
    LIPSYNC_DELAY = 0x40
    COMPRESSION = 0x41

    INCOMING_VIDEO_FORMAT = 0x42
    INCOMING_AUDIO_FORMAT = 0x43
    INCOMING_AUDIO_SAMPLERATE = 0x44

    SUB_STEREO_TRIM = 0x45  # Set/Request
    VIDEO_BRIGHTNESS = 0x46  # Set/Request
    VIDEO_CONTRAST = 0x47  # Set/Request
    VIDEO_COLOUR = 0x48  # Set/Request
    VIDEO_FILM_MODE = 0x49  # Set/Request
    VIDEO_EDGE_ENHANCEMENT = 0x4A  # Set/Request
    VIDEO_NOISE_REDUCTION = 0x4C  # Set/Request
    VIDEO_MPEG_NOISE_REDUCTION = 0x4D  # Set/Request
    ZONE_1_OSD_ON_OFF = 0x4E  # Set/Request
    VIDEO_OUTPUT_SWITCHING = 0x4F  # Set/Request
    VIDEO_OUTPUT_FRAME_RATE = 0x50  # Set/Request

    # 2.0 Commands
    INPUT_NAME = 0x20  # Set/Request
    FM_SCAN = 0x23
    DAB_SCAN = 0x24
    HEARTBEAT = 0x25
    REBOOT = 0x26


    @staticmethod
    def from_int(value: int):
        try:
            return CommandCodes(value)
        except ValueError:
            return value

class SourceCodes(enum.IntEnum):
    FOLLOW_ZONE_1 = 0x00
    CD = 0x01
    BD = 0x02
    AV = 0x03
    SAT = 0x04
    PVR = 0x05
    VCR = 0x06
    AUX = 0x08
    DISPLAY = 0x09
    FM = 0x0B
    DAB = 0x0C
    NET = 0x0E
    USB = 0x0F
    STB = 0x10
    GAME = 0x11

    @staticmethod
    def from_int(value: int):
        try:
            return SourceCodes(value)
        except ValueError:
            return value

    @staticmethod
    def from_bytes(value: bytes):
        return SourceCodes.from_int(int.from_bytes(value, 'big'))


class MenuCodes(enum.IntEnum):
    NONE = 0x00
    SETUP = 0x02
    TRIM = 0x03
    BASS = 0x04
    TREBLE = 0x05
    SYNC = 0x06
    SUB = 0x07
    TUNER = 0x08
    NETWORK = 0x09
    USB = 0x0A

    @staticmethod
    def from_int(value: int):
        try:
            return MenuCodes(value)
        except ValueError:
            return value

    @staticmethod
    def from_bytes(value: bytes):
        return MenuCodes.from_int(int.from_bytes(value, 'big'))


class DecodeMode2CH(enum.IntEnum):
    STEREO = 0x01
    DOLBY_PLII_IIx_MOVIE = 0x02
    DOLBY_PLII_IIx_MUSIC = 0x03
    DOLBY_PLII_IIx_GAME = 0x05
    DOLBY_PL = 0x06
    DTS_NEO_6_CINEMA = 0x07
    DTS_NEO_6_MUSIC = 0x08
    MCH_STEREO = 0x09

    @staticmethod
    def from_int(value: int):
        try:
            return DecodeMode2CH(value)
        except ValueError:
            return value

    @staticmethod
    def from_bytes(value: bytes):
        return DecodeMode2CH.from_int(int.from_bytes(value, 'big'))


class DecodeModeMCH(enum.IntEnum):
    STEREO_DOWNMIX = 0x01
    MULTI_CHANNEL = 0x02
    DOLBY_D_EX_OR_DTS_ES = 0x03
    DOLBY_PLII_IIx_MOVIE = 0x04
    DOLBY_PLII_IIx_MUSIC = 0x05

    @staticmethod
    def from_int(value: int):
        try:
            return DecodeModeMCH(value)
        except ValueError:
            return value

    @staticmethod
    def from_bytes(value: bytes):
        return DecodeModeMCH.from_int(int.from_bytes(value, 'big'))

class RC5Codes(enum.Enum):
    SELECT_STB = bytes([16, 1])
    SELECT_AV = bytes([16, 2])
    SELECT_TUNER = bytes([16, 3])
    SELECT_BD = bytes([16, 4])
    SELECT_GAME = bytes([16, 5])
    SELECT_VCR = bytes([16, 6])
    SELECT_CD = bytes([16, 7])
    SELECT_AUX = bytes([16, 8])
    SELECT_DISPLAY = bytes([16, 9])
    SELECT_SAT = bytes([16, 0])
    SELECT_PVR = bytes([16, 34])
    SELECT_USB = bytes([16, 18])
    SELECT_NET = bytes([16, 11])
    SELECT_DAB = bytes([16, 72])
    SELECT_FM = bytes([16, 54])
    INC_VOLUME = bytes([16, 16])
    DEC_VOLUME = bytes([16, 17])
    MUTE_ON = bytes([16, 119])
    MUTE_OFF = bytes([16, 120])
    DIRECT_MODE_ON = bytes([16, 78])
    DIRECT_MODE_OFF = bytes([16, 79])
    DOLBY_PLII_IIx_GAME = bytes([16, 102])
    DOLBY_PLII_IIx_MOVIE = bytes([16, 103])
    DOLBY_PLII_IIx_MUSIC = bytes([16, 104])
    MULTI_CHANNEL = bytes([16, 106])
    STEREO = bytes([16, 107])
    DOLBY_PL = bytes([16, 110])
    DTS_NEO_6_CINEMA = bytes([16, 111])
    DTS_NEO_6_MUSIC = bytes([16, 112])
    MCH_STEREO = bytes([16, 69])
    DOLBY_D_EX = bytes([16, 118])
    POWER_ON = bytes([16, 123])
    POWER_OFF = bytes([16, 124])
    FOLLOW_ZONE_1 = bytes([16, 20])

    MUTE_ON_ZONE2 = bytes([23, 4])
    MUTE_OFF_ZONE2 = bytes([23, 5])
    INC_VOLUME_ZONE2 = bytes([23, 1])
    DEC_VOLUME_ZONE2 = bytes([23, 2])

    SELECT_CD_ZONE2 = bytes([23, 6])
    SELECT_BD_ZONE2 = bytes([23, 7])
    SELECT_STB_ZONE2 = bytes([23, 8])
    SELECT_AV_ZONE2 = bytes([23, 9])
    SELECT_GAME_ZONE2 = bytes([23, 11])
    SELECT_AUX_ZONE2 = bytes([23, 13])
    SELECT_PVR_ZONE2 = bytes([23, 15])
    SELECT_FM_ZONE2 = bytes([23, 14])
    SELECT_DAB_ZONE2 = bytes([23, 16])
    SELECT_USB_ZONE2 = bytes([23, 18])
    SELECT_NET_ZONE2 = bytes([23, 19])
    POWER_ON_ZONE2 = bytes([23, 123])
    POWER_OFF_ZONE2 = bytes([23, 124])

SOURCECODE_TO_RC5CODE_ZONE1 = {
    SourceCodes.STB: RC5Codes.SELECT_STB,
    SourceCodes.AV: RC5Codes.SELECT_AV,
    SourceCodes.DAB: RC5Codes.SELECT_DAB,
    SourceCodes.FM: RC5Codes.SELECT_FM,
    SourceCodes.BD: RC5Codes.SELECT_BD,
    SourceCodes.GAME: RC5Codes.SELECT_GAME,
    SourceCodes.VCR: RC5Codes.SELECT_VCR,
    SourceCodes.CD: RC5Codes.SELECT_CD,
    SourceCodes.AUX: RC5Codes.SELECT_AUX,
    SourceCodes.DISPLAY: RC5Codes.SELECT_DISPLAY,
    SourceCodes.SAT: RC5Codes.SELECT_SAT,
    SourceCodes.PVR: RC5Codes.SELECT_PVR,
    SourceCodes.USB: RC5Codes.SELECT_USB,
    SourceCodes.NET: RC5Codes.SELECT_NET,
}

SOURCECODE_TO_RC5CODE_ZONE2 = {
    SourceCodes.STB: RC5Codes.SELECT_STB_ZONE2,
    SourceCodes.AV: RC5Codes.SELECT_AV_ZONE2,
    SourceCodes.DAB: RC5Codes.SELECT_DAB_ZONE2,
    SourceCodes.FM: RC5Codes.SELECT_FM_ZONE2,
    SourceCodes.BD: RC5Codes.SELECT_BD_ZONE2,
    SourceCodes.GAME: RC5Codes.SELECT_GAME_ZONE2,
    SourceCodes.CD: RC5Codes.SELECT_CD_ZONE2,
    SourceCodes.AUX: RC5Codes.SELECT_AUX_ZONE2,
    SourceCodes.PVR: RC5Codes.SELECT_PVR_ZONE2,
    SourceCodes.USB: RC5Codes.SELECT_USB_ZONE2,
    SourceCodes.NET: RC5Codes.SELECT_NET_ZONE2,
    SourceCodes.FOLLOW_ZONE_1: RC5Codes.FOLLOW_ZONE_1
}

class IncomingAudioFormat(enum.IntEnum):
    PCM = 0x00
    ANALOGUE_DIRECT = 0x01
    DOLBY_DIGITAL = 0x02
    DOLBY_DIGITAL_EX = 0x03
    DOLBY_DIGITAL_SURROUND = 0x04
    DOLBY_DIGITAL_PLUS = 0x05
    DOLBY_DIGITAL_TRUE_HD = 0x07
    DTS_96_24 = 0x08
    DTS_ES_MATRIX = 0x09
    DTS_ES_DISCRETE = 0x0A
    DTS_ES_MATRIX_96_24 = 0x0B
    DTS_ES_DISCRETE_96_24 = 0x0C
    DTS_HD_MASTER_AUDIO = 0x0D
    DTS_HD_HIGH_RES_AUDIO = 0x0E
    DTS_LOW_BIT_RATE = 0x0F
    DTS_CORE = 0x10
    PCM_ZERO = 0x13
    UNSUPPORTED = 0x14
    UNDETECTED = 0x15

    @staticmethod
    def from_int(value: int):
        try:
            return IncomingAudioFormat(value)
        except ValueError:
            return value

class IncomingAudioConfig(enum.IntEnum):
    """List of possible audio configurations."""
    MONO = 0x01
    CENTER_ONLY = 0x01
    STEREO_ONLY = 0x02
    # Incomplete list...

    @staticmethod
    def from_int(value: int):
        try:
            return IncomingAudioConfig(value)
        except ValueError:
            return value

@attr.s
class ResponsePacket():
    """Represent a response from device."""
    zn = attr.ib(type=int)
    cc = attr.ib(type=int)
    ac = attr.ib(type=int)
    data = attr.ib(type=bytes)

    @staticmethod
    def from_bytes(data: bytes) -> 'ResponsePacket':
        if len(data) < 6:
            raise InvalidPacket("Packet to short {}".format(data))

        if data[4] != len(data)-6:
            raise InvalidPacket("Invalid length in data {}".format(data))

        return ResponsePacket(
            data[1],
            CommandCodes.from_int(data[2]),
            AnswerCodes.from_int(data[3]),
            data[5:5+data[4]])

    def to_bytes(self):
        return bytes([
            *PROTOCOL_STR,
            self.zn,
            self.cc,
            self.ac,
            len(self.data),
            *self.data,
            *PROTOCOL_ETR
        ])

@attr.s
class CommandPacket():
    """Represent a command sent to device."""
    zn = attr.ib(type=int)
    cc = attr.ib(type=int)
    data = attr.ib(type=bytes)

    def to_bytes(self):
        return bytes([
            *PROTOCOL_STR,
            self.zn,
            self.cc,
            len(self.data),
            *self.data,
            *PROTOCOL_ETR
        ])

    @staticmethod
    def from_bytes(data: bytes) -> 'CommandPacket':
        if len(data) < 5:
            raise InvalidPacket("Packet to short {}".format(data))

        if data[3] != len(data)-5:
            raise InvalidPacket("Invalid length in data {}".format(data))

        return CommandPacket(
            data[1],
            CommandCodes.from_int(data[2]),
            data[4:4+data[3]])

async def _read_delimited(reader: asyncio.StreamReader, header_len: int) -> bytes:
    try:
        start = await reader.read(1)
        if start == PROTOCOL_EOF:
            _LOGGER.debug("eof")
            return None

        if start != PROTOCOL_STR:
            raise InvalidPacket("unexpected str byte {}".format(start))

        header = await reader.read(header_len-1)
        data_len = await reader.read(1)
        data = await reader.read(int.from_bytes(data_len, 'big'))
        etr = await reader.read(1)

        if etr != PROTOCOL_ETR:
            raise InvalidPacket("unexpected etr byte {}".format(etr))

        packet = bytes([*start, *header, *data_len, *data, *etr])
        return packet
    except TimeoutError as exception:
        raise ConnectionFailed() from exception
    except ConnectionError as exception:
        raise ConnectionFailed() from exception
    except OSError as exception:
        raise ConnectionFailed() from exception


async def _read_delimited_retried(reader: asyncio.StreamReader, header_len: int) -> bytes:
    while True:
        try:
            data = await _read_delimited(reader, header_len)
        except InvalidPacket as e:
            _LOGGER.warning(str(e))
            continue
        return data

async def _read_packet(reader: asyncio.StreamReader) -> ResponsePacket:
    data = await _read_delimited_retried(reader, 4)
    if not data:
        return None
    return ResponsePacket.from_bytes(data)


async def _read_command_packet(reader: asyncio.StreamReader) -> CommandPacket:
    data = await _read_delimited_retried(reader, 3)
    if not data:
        return None
    return CommandPacket.from_bytes(data)


async def _write_packet(writer: asyncio.StreamWriter,
                        packet: Union[CommandPacket,
                                      ResponsePacket]) -> None:
    try:
        data = packet.to_bytes()
        writer.write(data)
        await asyncio.wait_for(writer.drain(), _WRITE_TIMEOUT)
    except asyncio.TimeoutError as exception:
        raise ConnectionFailed() from exception
    except ConnectionError as exception:
        raise ConnectionFailed() from exception
    except OSError as exception:
        raise ConnectionFailed() from exception
